/*-
 * Copyright (c) 2005 Eugene Grosbein <eugen@grosbein.pp.ru>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] = "@(#)$Id: bwrite.c,v 1.1.1.1.2.2 2006/04/19 15:33:13 eugen Exp $";
#endif

/*
 * Implementation of buffering non-blocking writing.
 */

#include <sys/types.h>
#include <sys/queue.h>

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

int write_need = 0;		/* flag - does buffer need to be flushed? */

typedef struct {
  unsigned len;			/* memory buffer and its length		  */
  char *s;			/* s is not zero-terminated here	  */
} str;

struct buflist {		/* list of buffers			  */
  STAILQ_ENTRY(buflist) next;
  str buffer;
};

STAILQ_HEAD(WRBH, buflist);	
static struct WRBH head;	/* head of the list 			  */

/* RELENG_4 doesn't have STAILQ_FOREACH_SAFE */
#ifndef STAILQ_FOREACH_SAFE
#define	STAILQ_FOREACH_SAFE(var, head, field, tvar)			\
	for ((var) = STAILQ_FIRST((head));				\
	    (var) && ((tvar) = STAILQ_NEXT((var), field), 1);		\
	    (var) = (tvar))
#endif

/* Initialization of internal structures */
void write_init(void) {
  STAILQ_INIT(&head);
  write_need = 0;
}

/*
 * Try to flush buffer.
 * Write new strings s, if it is not NULL.
 * Keep it in the buffer, it output is busy.
 */
int write_out(char *s) {
  ssize_t rc;
  struct buflist *p, *temp;
  register size_t len;
  
  if (s) {
    if ((p = malloc(sizeof(*p))) == NULL)
      return -1;
    len = strlen(s);
    if ((p->buffer.s = malloc(p->buffer.len = len + 1)) == NULL)
      return -1;
    memcpy(p->buffer.s, s, len);
    p->buffer.s[len] = '\n';
    STAILQ_INSERT_TAIL(&head, p, next);
    write_need = 1;
  }

  STAILQ_FOREACH_SAFE(p, &head, next, temp) {
    len = p->buffer.len;
    if ((rc = write(STDOUT_FILENO, p->buffer.s, len)) == len) {
      free(p->buffer.s);
      STAILQ_REMOVE_HEAD(&head, next);
      free(p);
      continue;
    }
    if (rc > 0) {
      p->buffer.len -= rc;
      memcpy(p->buffer.s, p->buffer.s + rc, p->buffer.len);
    }
    else {
      if (errno == EPIPE)
        exit(0);
      if (errno == EAGAIN)
        return 0;
      else
         return -1;
    }
  }
 
  write_need = !(STAILQ_EMPTY(&head));
  return 0;
}
