/*-
 * Copyright (c) 1999 Eugene Grosbein <eugen@grosbein.pp.ru>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] = "@(#)$Id: ini.c,v 1.1.1.1.2.2 2006/04/19 15:46:26 eugen Exp $";
#endif

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "ini.h"

#define pifdbg(x) x

void free_inisections(ini_sections* current_sect)
{
  ini_sections* next_sect;
  ini_entry *current_entry, *next_entry;
  while(current_sect!=NULL) {         /* free section */
    next_sect=current_sect->next;      /* save next    */
    free(current_sect->section_name);

    current_entry=current_sect->first;  /* free parameters */
    while(current_entry!=NULL) {
      next_entry=current_entry->next;
      free(current_entry->name);
      free(current_entry->value);
      free(current_entry);
      current_entry=next_entry;
    }
    free(current_sect);
    current_sect=next_sect;
  }
}

ini_sections* parse_inifile(char* fname)
{
  FILE* f;
  ini_sections *res,*current_sect=NULL,*next_sect;
  ini_entry *current_par=NULL,*next_par;
  char buffer[INI_BUFFER];
  char *c, *sname, *name, *value;
  int s;

  res=NULL;

  if((f=fopen(fname,"rt"))==NULL) return NULL;
  while(!feof(f)) {
    while(1) {                         /* scanning file for section */
      if(fgets(buffer,INI_BUFFER,f)==NULL) {  /* if read error */
        if(feof(f)) break;
        pifdbg(puts("parse_inifile: read error");)
        free_inisections(res);
        errno=EIO;
        return NULL;
      }
      if((buffer[0]!='[') || ((c=strchr(buffer+1,']'))==NULL) ) continue; /* line ignored */
      s=(c-buffer);                      /* section found, allocating memory*/
      if((next_sect=malloc(sizeof(*next_sect)))==NULL) {  /* if no memory */
        pifdbg(puts("parse_inifile: no memory 1");)
        free_inisections(res);
        fclose(f);
        return NULL;
      }

      if((sname=malloc(s))==NULL) {              /* if no memory */
        pifdbg(puts("parse_inifile: no memory 2");)
        free(next_sect);
        free_inisections(res);
        fclose(f);
        return NULL;
      }

      if(res==NULL) res=next_sect;            /* for first section */
      else current_sect->next=next_sect;

      next_sect->next=NULL;
      next_sect->first=NULL;
      memcpy(sname,buffer+1,s-1);
      sname[s-1]=0;
      next_sect->section_name=sname;
      current_sect=next_sect;
      break;
    }

    while(1) {                        /* scanning file for parameters */
      if(fgets(buffer,INI_BUFFER,f)==NULL) {         /* if read error */
        if(feof(f)) break;
        pifdbg(printf("parse_inifile: read error inside section %s\n",current_sect->section_name);)
        free_inisections(res);
        errno=EIO;
        return NULL;
      }
      if(buffer[0]=='#') continue;                     /* comment is ignored */
      if((buffer[0]==0)||(buffer[0]=='\n')) break;      /* end of section   */
      if((c=strchr(buffer,'='))==NULL) continue; /* line that does not contain '=' is ignored */
      s=(c-buffer)+1;                                /* line found, allocating memory */

      if((next_par=malloc(sizeof(*next_par)))==NULL) {  /* if no memory */
        pifdbg(puts("parse_inifile: no memory 4");)
        free_inisections(res);
        fclose(f);
        return NULL;
      }

      if((name=malloc(s))==NULL) {
        pifdbg(puts("parse_inifile: no memory 5");)
        free(next_par);
        free_inisections(res);
        fclose(f);
        return NULL;
      }
      memcpy(name,buffer,s-1);
      name[s-1]=0;
      s=strlen(buffer)-s+1;
      if(buffer[strlen(buffer)-1]=='\n') s--;
      if((value=malloc(s))==NULL) {
        pifdbg(puts("parse_inifile: no memory 6");)
        free(next_par);
        free(name);
        free_inisections(res);
        fclose(f);
        return NULL;
      }

      memcpy(value,c+1,s-1);
      value[s-1]=0;

      if(current_sect->first==NULL) current_sect->first=next_par; /* for first section */
      else current_par->next=next_par;

      next_par->name=name;
      next_par->value=value;
      next_par->next=NULL;

      current_par=next_par;
    }
  }
  fclose(f);
  return res;
}

ini_sections* get_section(ini_sections* ini, char* name)
{
  ini_sections* current;

  current=ini;
  while(current!=NULL) 
   if(strcmp(current->section_name,name)==0) break;
   else current=current->next;

  return current;
}

ini_entry* get_entry(ini_sections* sect, char* name)
{
   ini_entry* current;

  current=sect->first;
  while(current!=NULL) 
   if(strcmp(current->name,name)==0) break;
   else current=current->next;

  return current;
}

char* get_parameter(ini_sections* ini, char* sname, char* pname, char* def)
{
  ini_sections* sect;
  ini_entry* par;

  if((sect=get_section(ini,sname))==NULL) return def;
  if((par=get_entry(sect,pname))==NULL) return def;
  else return par->value;
}

/*
int main(void)
{
  ini_sections* ini;
  ini_sections *current_s;
  ini_entry *current;
  ini=parse_inifile("test.ini");
  if(ini==NULL) {
    switch(errno) {
    case ENOFILE: puts("File not found"); break;
    }
    return errno;
  }
  current_s=ini;
  while(current_s!=NULL) {
    printf("[%s]\n",current_s->section_name);
    current=current_s->first;
    while(current!=NULL) {
      printf("%s=%s\n",current->name,current->value);
      current=current->next;
    }
    puts("");
    current_s=current_s->next;
  }

  free_inisections(ini);
  return 0;
}
*/

