/*-
 * Copyright (c) 2005 Eugene Grosbein <eugen@grosbein.pp.ru>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] = "@(#)$Id: ircdec.c,v 1.1.1.1.2.3 2006/04/19 16:31:00 eugen Exp $";
#endif

/*
 * This program reads data from InfraRed Dongle and translates it
 * to button names of InfraRed Control Unit.
 *
 * Tested with Tekram IR-210B and several IRC Units.
 */

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <termios.h>
#include <unistd.h>

#include "common.h"
#include "bwrite.h"
#include "ini.h"
#include "irc.h"
#include "mathmodel.h"

#define IRCCONF "/usr/local/etc/ircdec.conf"

static ircdef irc;		/* InfraRed Control unit description	 */
static char* device;		/* serial device file name 		 */
static int  fd,			/* file descriptor for device		 */
	    kq;			/* kqueue() descriptor	 		 */
static speed_t speed;		/* speed value from config file		 */
static unsigned reset = 0;	/* reset period (sec.) for idle device	 */

extern unsigned max_interval;   /* used in debug/training mode only      */
int debug = 0;			/* debug level 				 */

static char *pfname = NULL;	/* atexit()-function removes PID file	 */
static struct termios tistd;	/* and restores terminal state		 */
int restore_termios = 0;

static void	cleanup(void);
static void	flush_out(char *);
static void	initialize(char *, char*);
static int	interrupted(int, struct kevent *);
static void	listen_out(int);
static void	opendev(void);
static int	read_raw(void*, const ssize_t);

/*
 * Open and setup serial device
 */
static void opendev(void) {
  struct kevent ev_read;	/* EV_READ filter for device		 */
  struct termios ti;		/* to set speed, flow control and CLOCAL */

  /* open irc device non-blocking, otherwise open() may block until CD */
  if ((fd = open(device, O_RDONLY | O_NONBLOCK)) == -1)
	err(EX_IOERR, "cannot open %s", device);

  /* setup serial device parameters */
  if (tcgetattr(fd, &ti) < 0)
	err(EX_IOERR, "tcgetattr");
	
  /* set speed */
  if (cfsetospeed(&ti, speed) < 0)
	err(EX_IOERR, "cfsetospeed");
  if (cfsetispeed(&ti, speed) < 0)
	err(EX_IOERR, "cfsetospeed");

  /* disable software and hardware flow control, disable modem control */
  /* disable most of input translations */
  ti.c_iflag &= ~(INLCR | IGNCR | ICRNL | IXON | IXOFF | IXANY | IMAXBEL |
		  IGNBRK | BRKINT);
  ti.c_cflag &= ~(CCTS_OFLOW | CRTS_IFLOW | CLOCAL);
  ti.c_lflag &= ~(ICANON | ISIG | IEXTEN);
  if (tcsetattr(fd, TCSANOW, &ti) < 0)
	err(EX_IOERR, "tcsetattr: check speed(%u) value", speed);

  /* clear O_NONBLOCK flag, we'll use blocking read */
  if (fcntl(fd,F_SETFL,fcntl(fd,F_GETFL) & !O_NONBLOCK) < 0)
	errx(EX_IOERR, "fcntl: cannot clear O_NONBLOCK on %s", device);

  /* set O_NONBLOCK flag on STDOUT */
  if (fcntl(STDOUT_FILENO, F_SETFL,
	    fcntl(STDOUT_FILENO,F_GETFL) | O_NONBLOCK) < 0)
	errx(EX_IOERR, "fcntl: cannot set O_NONBLOCK on stdout");

  /* initialize kqueue */
  if ((kq = kqueue()) < 0)
    err(EX_OSERR, "kqueue");
    
  /* watch for data from device */
  EV_SET(&ev_read, fd, EVFILT_READ, EV_ADD, 0, 0, 0);
  if (kevent(kq, &ev_read, 1, NULL, 0, NULL) < 0)
    err(EX_OSERR, "kevent");
}

/*
 * Complete initialization
 */
static void
initialize(char *config, char *ircsect)
{
  char	*bd,		/* button in a config file			*/
  	*d,		/* iterator over an array of buttons		*/
	*s,		/* temporary pointer for a string		*/
	*rest;		/* another temporary storage for a string 	*/

  ini_sections	*ini;	 /* parsed configuration file		 */
  buttondef	*button; /* a button definition is built using config	*/
  ini_entry	*ie;	 /* an entry of a section of a configuration	*/

  unsigned long l;	/* temporary storage for an unsigned long	*/
  int	buttons,	/* buttons number of irc			*/
	line;		/* directive number of config section		*/

  register unsigned char byte;	/* byte value of two hex chars		*/

  /* load and parse config file */
  if ((ini = parse_inifile(config))== NULL)
    err(EX_DATAERR, "cannot parse %s", config);

/* sanity checks for values got from config file */
#define GETCONF(var,sect,entr) 			\
  do {						\
    var = get_parameter(ini, sect, entr, NULL);	\
    if (var == NULL)				\
      errx(EX_DATAERR, "%s: %s not found in the section %s", config, entr, sect); \
  } while(0)

#define GETUINTCONF(var,type,sect,entr)		\
  do {						\
    GETCONF(s,sect,entr);			\
    l = strtoul(s, &rest, 10);			\
    if (*rest != '\0' || l > UINT_MAX/2)	\
      errx(EX_DATAERR, "%s: %s(%s) must be positive integer less than %u in the section %s", \
	   config, entr, s, UINT_MAX/2, sect);	\
    var = (type) l;				\
  } while(0)

  /* extract basic parameters from the configuration */
  /* and perform sanity checks */
  
  /* get device file name from [comms] */
  GETCONF(device, "comms", "device");	
 
  /* get speed from [comms] */
  GETUINTCONF(speed, speed_t, "comms", "speed");
  
  /* get irc section name from [comms] */
  if (!ircsect)
    GETCONF(ircsect, "comms", "irc");	

  /* now get values from irc section */
  /* get irc name */
  GETCONF(irc.name, ircsect,"name");

  /* get button timeout (in microseconds) */
  GETUINTCONF(irc.timeout, unsigned, ircsect,"timeout");

  /* get maximum bit error allowed */
  GETUINTCONF(irc.maxber, unsigned, ircsect,"maxber");

  /* XXX - nearly the same code as GETUINTCONF */
  if ((s = get_parameter(ini, ircsect, "speed", NULL)) != NULL) {
    l = strtoul(s, &rest, 10);
    if (*rest != '\0' || l > UINT_MAX/2)
      errx(EX_DATAERR, "%s: speed(%s) must be positive integer less than %u in the section %s",
	 config, s, UINT_MAX/2, ircsect);
    speed = (speed_t) l;
  }

  if (irc.jred == -1)
    irc.jred = irc.timeout * 5;

  /* XXX - two passes over ircsect. Should replace with one pass */
  /* compute number of buttons */
  for (buttons = 0, ie = get_section(ini,ircsect)->first;
       ie != NULL; ie = ie->next)
    if ((strcmp(ie->name,"button") == 0) || (strcmp(ie->name,"repeat") == 0))
	buttons++;

  irc.max_button = buttons - 1;

  /* allocate memory for the button array */
  if ((irc.button = button = malloc(buttons*sizeof(*button))) == NULL)
    err(EX_OSERR, "malloc");

  /* parse button definitions */
  for (ie = get_section(ini,ircsect)->first, line = 1;
       ie != NULL; ie = ie->next, line++) {
    if (strcmp(ie->name, "repeat") == 0)
      button->flags = BUTTON_REPEAT;	/* not real button but REPEAT sign */
    else {
      if (strcmp(ie->name, "button") != 0)
        continue;
      button->flags = 0;
    }

    /* do not write into parsed structure of config, use copy instead */
    if ((bd = strdup(ie->value)) == NULL)
	err(EX_OSERR, "malloc");

    /* label and bytestring of button must be separated with space */
    if ((button->label = strtok(bd," \t")) == NULL)
      errx(EX_DATAERR, "%s(line %d): cannot parse %s\n",
    			config, line, ie->value);
    
    /* store label to the buttondef structure */
    if ((button->label = strdup(button->label)) == NULL)
      err(EX_OSERR, "malloc");

    /* s points to the start of the bytestring (in hex) of a button */
    if ((s = strtok(NULL, " \t")) == NULL)
      errx(EX_DATAERR, "%s(line %d): cannot parse %s\n",
        config, line, ie->value);

    /* calculate number of bytes needed to store bytestring */
    if ((button->len = strlen(s)) % 2) {
      button->len = button->len/2 + 1;
      *(--s) = '0'; /* add leading zero for first byte if it's missing */
    }
    else
      button->len /= 2;
    
    /* memory for the bytestring */
    if ((button->definition = d = malloc(button->len)) == NULL)
      err(EX_OSERR, "malloc");

    /* now convert hex characters into bytes */    
    while (*s) {
      if (*s >= '0' && *s <= '9')
        byte = (*s - '0') << 4;
      else if (*s >= 'a' && *s <= 'f')
        byte = (*s - 'a' + 10) << 4;
      else if (*s >= 'A' && *s <= 'F')
        byte = (*s - 'A' + 10) << 4;
      else
        errx(EX_DATAERR, "%s: value of %s is not hex string\n",
	     config, ie->value);

      s++;
      if (*s >= '0' && *s <= '9')
        byte += (*s - '0');
      else if (*s >= 'a' && *s <= 'f')
        byte += (*s - 'a' + 10);
      else if (*s >= 'A' && *s <= 'F')
        byte += (*s - 'A' + 10);
      else
        errx(EX_DATAERR, "%s: value of %s is not hex string\n",
	     config, ie->value);
      s++;

      *(d++) = byte;
    }
    /* all data is stored into structures, we don't need a copy anymore */
    free(bd);
    button++;
  }

  /* sort array of buttons to speedup lookups later */
  qsort(irc.button, buttons, sizeof(irc.button[0]), &compare);
  
  /* initialize write buffering subsystem */
  signal(SIGPIPE, SIG_IGN);
  write_init();

  if (debug) {
    /* set NOKERNINFO  */
    if (isatty(STDIN_FILENO)) {
      struct termios ti;

      if (tcgetattr(STDIN_FILENO, &ti) < 0)
	err(EX_IOERR, "tcgetattr(stdin)");

      tistd = ti;
      ti.c_lflag |= NOKERNINFO;
      if (tcsetattr(fd, TCSANOW, &ti) < 0)
	err(EX_IOERR, "tcsetattr(stdin)");
      restore_termios = 1;
    }
    siginterrupt(SIGINFO, 0);
    signal(SIGINFO, sighandler);
  }

  /* open device and initialize kqueue() */
  opendev();
}

/*
 * Returns true if kqueue() was interrupted and should be restarted.
 * Ctrl-T may be used to show stats for training mode.
 * Checks for EOF and other unexpected errors and halts.
 */
static int
interrupted(int e, struct kevent *ev) {
  if (e == 0)
    return 0;
  if (e == -1) {
    if (errno == EINTR) {
      if (debug)
      	print_stats();
      return 1;
    }
    else
      err(EX_IOERR,"kevent");
  }
  for(; e > 0 ; e--, ev++) {
    if (ev->ident == fd) {
      if (ev->flags & EV_EOF)
        exit(0);
      if (ev->flags & EV_ERROR)
        exit(EX_IOERR);
      return 0;
    }
  }
  return 0;
}

/*
 * Adds (flag==1) or removes (flag==0) EVFILT_WRITE for STDOUT.
 */
void listen_out(int flag) {
  struct kevent ev_write;

  EV_SET(&ev_write, STDOUT_FILENO, EVFILT_WRITE,
	 flag ? EV_ADD : EV_DELETE, 0, 0, 0);
  if ((kevent(kq, &ev_write, STDOUT_FILENO, NULL, 0, NULL) < 0) &&
      (errno != ENOENT)) {
    if (errno == EPIPE)
      exit(0);
    else
      err(EX_OSERR, "listen_out(%d): kevent", flag);
  }
}

/*
 * Writes out buffer and one new string, if possible.
 * Or adds new string to the write buffer.
 * Enables (for non-empty buffer) or disables kqueue() filter for STDOUT.
 */
void flush_out(char *s) {
  write_out(s);
  if (write_need)
    listen_out(1);
  else
    listen_out(0);
}

/*
 * Constructs raw bytestring from chunks of data read from device,
 * writes to the buffer b upto s bytes.
 */
static int
read_raw(void *b, const ssize_t s)
{
  struct kevent ev[2];		/* working area for kqueue()		*/
  struct timespec tm;		/* to specify the timeout for kevent()	*/
  struct timeval last, was;	/* for timestamps			*/
  unsigned char *buf = b;		/* sliding pointer to unfilled part	*/
				/* of the buffer 			*/

  ssize_t n,			/* value returned by read()		*/
	  rest = s;		/* free bytes in the buffer		*/
  unsigned interval = 0;	/* measured timeout for a button	*/
  int e;			/* error state for kevent()		*/
  int flush_later = 0;		/* flag: is stdout ready
				   and buffer not empty?		*/
  
  if (s <= 0) return 0;		/* sanity check */

  if (reset) {			/* reset device if it is idle?		*/
    tm.tv_sec = reset;
    tm.tv_nsec = 0;
  }

  /* wait for the first chunk of data  */
  for (;;) {
    /* Sleep until data is here or STDOUT is ready to receive old labels */
    e = kevent(kq, NULL, 0, &ev[0], 2, reset ? &tm : NULL);
    if (e == 0) { /* if reset period was supplied and device is idle, reset */
      close(fd);
      opendev();
      continue;
    }

    /* check for EINTR and errors */
    if (interrupted(e, &ev[0]))
        continue;
    if (ev[0].ident == STDOUT_FILENO) {
      if (e == 1) {		/* only one event: STDOUT became ready   */
        flush_out(NULL);	/* flush write buffer, no new labels yet */
        continue;
      }
      else {			/* two events: STDOUT is ready and	 */
        flush_later = 1;	/* new data arrived; read new data first */
        ev[0] = ev[1];
      }
    }
    /* there is a byte to read, make timestamp for this moment */
    gettimeofday(&was, NULL);
    last = was;
    if ((n = read(fd, buf, rest)) <= 0)
      err(EX_IOERR, "read");	/* should not happen */
    buf  += n;
    rest -= n;
    break; /* proceed to the main loop for other chunks */
  }

  tm.tv_sec = 0;
  tm.tv_nsec = 10000000;	/* XXX: hardcoded HZ (10000000 nanosec) */

  /* stop looking at STDOUT */
  listen_out(0);

  /* while there is a space in a buffer */
  while (rest>0) {
  /* sleep until timeout elapsed */
    while ((rest>0) && (e = kevent(kq, NULL, 0, &ev[0], 1, &tm)) > 0) {
      if (interrupted(e, &ev[0]))
        continue;
      if ((n = read(fd, buf, rest)) <= 0)
        err(EX_IOERR, "read"); /* should not happen */
      else {
        buf  += n;
        rest -= n;
	if (debug) {
          gettimeofday(&last, NULL);
  	  interval = (last.tv_sec-was.tv_sec)*1000000+last.tv_usec-was.tv_usec;
	} 
      }
    }

    if (interrupted(e, &(ev[0])))
      continue;

    gettimeofday(&last, NULL);

    if (debug > 1) {
      fprintf(stderr, "time: %lu/%u\t\t\b\b\b\b\b",
    		(last.tv_sec-was.tv_sec)*1000000+last.tv_usec-was.tv_usec,
		irc.timeout);
      printbuf(b, s-rest);
      fprintf(stderr, "\n");
    }
    /* stop waiting for additional bytes if button timeout has elapsed */
    if (irc.timeout <=
	(last.tv_sec-was.tv_sec)*1000000+last.tv_usec-was.tv_usec)
      break;
  }
  
  if (debug) {
    if (max_interval < interval)
      max_interval = interval;
    update_stats(b, s-rest);
    fprintf(stderr, "time: %5lu/%05u=%5.1f%%\t\b\b\b\b\b",
      (last.tv_sec-was.tv_sec)*1000000+last.tv_usec-was.tv_usec,
      interval,
      interval ?
       ((last.tv_sec-was.tv_sec) * 1e6 + (last.tv_usec-was.tv_usec)) * 100 / interval :
       99999);
    printbuf(b, s-rest);
  }
  if (flush_later)
    flush_out(NULL);

  /* return the size of the bytestring */
  return s-rest;
}

void cleanup(void) {
  if (pfname != NULL)
    unlink(pfname);
  if (restore_termios)
    tcsetattr(fd, TCSANOW, &tistd);	/* restore NOKERNINFO state */
}

#define OPTIONS "c:df:j:p:r:u"
#define USAGE "Usage: %s [-c control] [-d] [-f config] [-j delay] [-p pidfile]" \
		" [-r delay] [-u]\n"
int
main(int argc, char* argv[])
{
  unsigned char buf[MAXBUF];	/* space for data coming from device	 */
  buttondef *button,		/* decoded button			 */
	    *prevbutton = NULL,
	    *abutton;
  struct timeval
    last_time = { 0, 0 }, now;	/* working area for jitter reduction	 */
    
  char	*config = IRCCONF; 	/* name of config file, override with -f */
  char  *ircsect;		/* name of section describing IR Control */
  int	ch, show_unknown = 0;
  FILE	*pidfile;
  int	jred;			/* cache irc.jred in a local var	 */
  int repeat;

  irc.jred = -1;

  config = getenv("IRCCONF");
  if (!config)
    config = IRCCONF;
  ircsect = getenv("IRCU");

  atexit(&cleanup);

  while ((ch = getopt(argc, argv, OPTIONS)) != -1)
    switch (ch) {
    case 'c':
	ircsect = optarg;
	break;
    case 'd':
	debug++;
	show_unknown = 1;
	irc.jred = 0;
	break;
    case 'f':
	config = optarg;
	break;
    case 'j':
	irc.jred = atoi(optarg) * 1000;
	break;
    case 'p':
	pfname = optarg;
	if ((pidfile = fopen(pfname, "w")) == NULL)
	  err(EX_IOERR, "cannot open %s", pfname);
	setbuf(pidfile, NULL);
	if (fprintf(pidfile, "%u\n", getpid()) < 0)
	  err(EX_IOERR, "error writing %s", pfname);
	if (fclose(pidfile) < 0)
	  err(EX_IOERR, "cannot close %s", pfname);
	break;
    case 'r':
	reset = atoi(optarg);
	break;
    case 'u':
	show_unknown = 1;
	break;
    default:
	errx(EX_USAGE,USAGE,argv[0]);
	/* NOTREACHED */
  }

  initialize(config, ircsect);
  jred = irc.jred;

  for (;;) {			/* main endless loop		*/
    for (;;) {			/* loop for jitter reduction	*/
      button = find_button(buf, read_raw(buf,sizeof(buf)), irc);
      if (button == NULL) {	/* decode failure */
        prevbutton = NULL;	/* if next button will be repeat sign, ignore */
	if(show_unknown)
	  flush_out("UNKNOWN");
	continue;
      }
      if ((repeat = (button->flags & BUTTON_REPEAT))) {
        if (!prevbutton)	/* repeat of undecoded button - ignore */
	  continue;
      }

      abutton = repeat ? prevbutton : button;
      if (jred) {		/* jitter reduction is requested */
        gettimeofday(&now, NULL);
	/* do not ignore mouse events;
	 * ignore button if previous was decoded
	 * and this time is the same (repeat or just same button)
	 * and time interval is too small
	 */
        if (prevbutton && last_time.tv_sec && 
	     (repeat || (strcmp(prevbutton->label, button->label) == 0)) &&
	     ( ((now.tv_sec-last_time.tv_sec) * 1000000 + 
	         now.tv_usec-last_time.tv_usec) < jred
	     )
	   ) {
	  continue;
	}
	last_time = now;
      }
        flush_out(button->label);
      if (!repeat)
        prevbutton = button;

    }
  }

/* NOTREACHED */
/*close(fd);*/

  return 0;
}
