/*-
 * Copyright (c) 2005 Eugene Grosbein <eugen@grosbein.pp.ru>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] = "@(#)$Id: mathmodel.c,v 1.1.1.1.2.2 2006/04/19 16:04:58 eugen Exp $";
#endif

/*
 * Math area. Detection logic and misc functions.
 * No global variables and side effects here (except of debug output).
 */
 
#include <sys/types.h>
#include <limits.h>
#include <stdio.h>

#include "irc.h"

extern int debug;

/*
 * The comparison function for qsort(3)
 */
int
compare(const void *b1, const void *b2) {
  size_t l1 = ((buttondef*)b1) -> len, l2 = ((buttondef*)b2) -> len;
  
  if (l1 == l2)
    return 0;
  else if (l1 > l2)
    return 1;
  else return -1;
}

/*
 * Metric is a number of non-equal bits in bitstrings b1 and b2,
 * both are len bytes in length.
 */
static
unsigned
metric(const void *b1, const void *b2, ssize_t len)
{
  /* bc[b] equals to number of non-zero bits in a byte b */
  static const unsigned char bc[] = {
    0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
  };

  const unsigned char *buf1 = b1, *buf2 = b2;
  register int i;		/* loop counter */
  register unsigned m = 0;	/* metric value */

  for (i = len-1; i >= 0; i--, buf1++, buf2++) {
    m += bc[(*buf1) ^ (*buf2)];
  }

  return m;
}

/*
 * Find button for bitstring buf.
 * Return NULL if button not found.
 */
buttondef *
find_button(const unsigned char* buf, const ssize_t len, const ircdef irc)
{
  buttondef	*button = irc.button,	/* first button in array	*/
		*bestbut = NULL;	/* most suitable guess		*/
  unsigned	 distance = UINT_MAX,	/* metric of a button		*/
		 best = UINT_MAX;	/* metric of best guess		*/
  register int	 i;			/* loop counter 		*/
  
  for (i = irc.max_button; i >= 0; i--, button++) {
    if (button->len < len)	/* skip bitstrings with wrong length    */
      continue;
    if (button->len > len)	
      break;

    distance = metric(buf, button->definition, len);
    if (distance < best) {
      bestbut = button;
      if ((best = distance) == 0) /* stop search when exact match found */
        break;
    }
  }
  if (debug)
    fprintf(stderr, "metric %u ", best);
  if (best > irc.maxber) /* no similar button found :-( */
    return NULL;

  return bestbut;	/* our guess is good enough */
}
